import vcf
import pyranges
from .json_generator import _get_fhir_json
import logging
import sys
from .common import *
general_logger = logging.getLogger('vcf2fhir.general')
"""Converter for a VCF version >4.1 file."""


class Converter(object):
    """Creates a new Converter Object to convert a VCF file.

    Parameters
    ----------

    **vcf_filename** (required): Path to a text-based or bgzipped VCF
    file. Valid path and filename without whitespace must be provided.
    VCF file must conform to VCF Version 4.1 or later. FORMAT.GT must be
    present. Multi-sample VCFs are allowed, but only the first sample
    will be converted.
    has_tabix (required if VCF file is bgzipped): Set to 'True' if
    there is a tabix index. Tabix file must have the same name as the
    bgzipped VCF file, with a '.tbi' extension, and must be in the same
    folder.

    **ref_build** (required): Genome Reference Consortium genome
    assembly to which variants in the VCF were called. Must be one of
    'GRCh37' or 'GRCh38'.

    **patient_id** (optional): Supplied patient ID is inserted into
    generated FHIR output. Alphanumeric string without whitespace. if \
    not provided,
    header of first sample column is used.

    **Conversion region** (optional): Subset of the VCF file to be
    converted into FHIR. If absent, the entire VCF file is converted. Can
    be supplied as either a parameter (conv_region_dict) or as a BED
    file (conv_region_filename):

       **conv_region_dict** : Array of regions (e.g. '{"Chromosome":
       ["X", "X", "M"],"Start": [50000, 55000, 50000],"End": [52000,
       60600, 60025]}'). Values for Chromosome must align with values in
       VCF #CHROM field. Ranges must be `0-based \
       <https://www.biostars.org/p/84686/>`_
       (or 0-start, half-open) and based on GRCh37 or GRCh38 \
       reference sequences.

       **conv_region_filename**: Valid path and filename without
       whitespace must be provided. Must be a valid BED file with first 3
       columns: <chr> <start> <stop>. Values in <chr> field must align
       with values in VCF #CHROM field. Ranges must be based on GRCh37 or
       GRCh38 reference sequences. Note that BED files are
       `0-based <https://www.biostars.org/p/84686/>`_ (or 0-start,
       half-open) whereas VCF files and FHIR output are 1-based (or
       1-start, fully-closed).

    **region_studied_filename** (optional): Subset of patient's genome
    that was studied in the generation of the VCF file. Valid path and
    filename without whitespace must be provided. Must be a valid BED
    file, with first 3 columns: <chr> <start> <stop>. Values in <chr>
    field must align with values in VCF #CHROM field. Ranges must be
    based on GRCh37 or GRCh38 reference sequences. Note that BED files
    are `0-based <https://www.biostars.org/p/84686/>`_ (or 0-start,
    half-open) whereas VCF files and FHIR output are 1-based (or 1-start,
    fully-closed).

    **nocall_filename** (optional): Subset of studied region that is
    deemed noncallable. Valid path and filename without whitespace must
    be provided. Must be a valid BED file, with first 3 columns: <chr>
    <start> <stop>. Values in <chr> field must align with values in VCF
    #CHROM field. Ranges must be based on GRCh37 or GRCh38 reference
    sequences. Note that BED files are `0-based \
    <https://www.biostars.org/p/84686/>`_
    (or 0-start, half-open) whereas VCF files and FHIR output are \
    1-based (or 1-start,
    fully-closed).

    **ratio_ad_dp** (optional)(default value = 0.99): This ratio \
    determine whether to assign Homoplasmic or Heteroplasmic
     If allelic depth (FORMAT.AD) / read depth (FORMAT.DP) is \
     greater than ratio_ad_dp then allelic state is
      homoplasmic; else heteroplasmic.
    Returns
    -------

    Object
    An Instance of Converter that helps to convert vcf file.

    """

    def __init__(
            self,
            vcf_filename=None,
            ref_build=None,
            patient_id=None,
            has_tabix=False,
            conv_region_filename=None,
            conv_region_dict=None,
            region_studied_filename=None,
            nocall_filename=None,
            ratio_ad_dp=0.99):

        super(Converter, self).__init__()
        if not (vcf_filename):
            raise Exception('You must provide vcf_filename')
        if not ref_build or ref_build not in ["GRCh37", "GRCh38"]:
            raise Exception(
                'You must provide build number ("GRCh37" or "GRCh38")')
        if nocall_filename and not region_studied_filename:
            raise Exception(
                ("Please also provide region_studied_filename " +
                 "when nocall_filename is provided"))
        self.vcf_filename = vcf_filename
        try:
            self._vcf_reader = vcf.Reader(filename=vcf_filename)
        except FileNotFoundError:
            raise
        except BaseException:
            self._generate_exception("Please provide valid  'vcf_filename'")
        if not patient_id:
            patient_id = self._vcf_reader.samples[0]
        if nocall_filename:
            try:
                self.nocall_region = pyranges.read_bed(nocall_filename)
            except FileNotFoundError:
                raise
            except BaseException:
                self._generate_exception(
                    "Please provide valid  'nocall_filename'")
        else:
            self.nocall_region = pyranges.PyRanges()
        if conv_region_filename:
            try:
                self.conversion_region = pyranges.read_bed(
                    conv_region_filename)
            except FileNotFoundError:
                raise
            except BaseException:
                self._generate_exception(
                    "Please provide valid 'conv_region_filename'")
        elif conv_region_dict:
            try:
                self.conversion_region = pyranges.from_dict(conv_region_dict)
            except FileNotFoundError:
                raise
            except BaseException:
                self._generate_exception(
                    "Please provide valid 'conv_region_dict'")
        else:
            self.conversion_region = None
        if region_studied_filename:
            try:
                self.region_studied = pyranges.read_bed(
                    region_studied_filename)
            except FileNotFoundError:
                raise
            except BaseException:
                self._generate_exception(
                    "Please provide valid 'region_studied_filename'")
        else:
            self.region_studied = None

        if not validate_has_tabix(has_tabix):
            raise Exception("Please provide a valid 'has_tabix'")

        if not validate_ratio_ad_dp(ratio_ad_dp):
            raise Exception("Please provide a valid 'ratio_ad_dp'")

        self.ratio_ad_dp = ratio_ad_dp
        self.has_tabix = has_tabix
        self.patient_id = patient_id
        self.ref_build = ref_build
        self.nocall_filename = nocall_filename
        self.conv_region_filename = conv_region_filename
        general_logger.info("Converter class instantiated successfully")

    def convert(self, output_filename='fhir.json'):
        """ Generates HL7 FHIR Genomics format data as output_filename \
        or fhir.json if it is not provided

        Parameters
        ----------
        output_filename:
            Path to output fhir json.

        """
        general_logger.info("Starting VCF to FHIR Conversion")
        _get_fhir_json(
            self._vcf_reader,
            self.ref_build,
            self.patient_id,
            self.has_tabix,
            self.conversion_region,
            self.region_studied,
            self.nocall_region,
            self.ratio_ad_dp,
            output_filename)
        general_logger.info("Completed VCF to FHIR Conversion")

    def _generate_exception(self, msg):
        general_logger.error(msg, exc_info=True)
        raise Exception(msg, sys.exc_info)



import doctest
import unittest
import os
import vcf2fhir
import json
from os.path import join, dirname
import shutil
import logging

suite = doctest.DocTestSuite(vcf2fhir)


def _get_uids_map(fhir_json):
    obv_ids = []
    result_ids = []
    for index, obv in enumerate(fhir_json['contained']):
        id = obv['id']
        obv_ids.append(f'#{id}')
        fhir_json['contained'][index]['id'] = ''

    for index, result in enumerate(fhir_json['result']):
        result_ids.append(result['reference'])
        fhir_json['result'][index]['reference'] = ''
    fhir_json['id'] = ''
    return {'obv_ids': obv_ids, 'result_ids': result_ids}


def _validate_phase_rel(fhir_json, map_variant_index):
    for index_seq, list_index_var in map_variant_index.items():
        for index, ref in enumerate(
                fhir_json['contained'][index_seq]['derivedFrom']):
            variant_id =\
                fhir_json['contained'][list_index_var[index]]['id']
            if not ref['reference'] == f'#{variant_id}':
                return False
            fhir_json[
                        'contained'
                    ][
                        index_seq
                    ][
                        'derivedFrom'
                    ][
                        index
                    ][
                        'reference'
                    ] = ''
    return True


def _compare_actual_and_expected_fhir_json(
        self,
        output_filename,
        expected_output_filename,
        dict=None,
        is_null=False):
    with open(output_filename) as output_file,\
            open(expected_output_filename) as expected_output_file:
        actual_fhir_json = json.load(output_file)
        # Validate the passed sequence relationship
        if not is_null:
            self.assertEqual(_validate_phase_rel(
                actual_fhir_json, dict), True)
        # Validate: list of observation uids and list of result uids same.
        # Also set the uids to '' to avoid guid comparison in next step
        map_ids = _get_uids_map(actual_fhir_json)
        self.assertEqual(map_ids['obv_ids'], map_ids['result_ids'])
        actual_fhir_json['issued'] = ''
        # Finally, check if the acutal json after removing all uids is
        # same as expected json
        expected_fhir_json = json.load(expected_output_file)
        self.assertEqual(actual_fhir_json, expected_fhir_json)


class TestVcf2FhirInputs(unittest.TestCase):

    def test_required_vcf_filename(self):
        with self.assertRaises(Exception) as context:
            vcf2fhir.Converter()
        self.assertTrue(
            'You must provide vcf_filename' in str(context.exception))

    def test_invalid_vcf_filename(self):
        self.assertRaises(FileNotFoundError,
                          vcf2fhir.Converter, *['Hello', 'GRCh38'])

    def test_required_ref_build(self):
        with self.assertRaises(Exception) as context:
            vcf2fhir.Converter(os.path.join(
                os.path.dirname(__file__), 'vcf_example1.vcf'))
        self.assertEqual(
            'You must provide build number ("GRCh37" or "GRCh38")', str(
                context.exception))

    def test_invalid_ref_build(self):
        with self.assertRaises(Exception) as context:
            vcf2fhir.Converter(os.path.join(
                os.path.dirname(__file__), 'vcf_example1.vcf'), 'b38')
        self.assertEqual(
            'You must provide build number ("GRCh37" or "GRCh38")', str(
                context.exception))

    def test_valid_ref_build_37(self):
        o_vcf_2_fhir = vcf2fhir.Converter(os.path.join(
            os.path.dirname(__file__), 'vcf_example1.vcf'), 'GRCh37')
        self.assertEqual(type(o_vcf_2_fhir), vcf2fhir.Converter)

    def test_valid_ref_build_38(self):
        o_vcf_2_fhir = vcf2fhir.Converter(os.path.join(
            os.path.dirname(__file__), 'vcf_example1.vcf'), 'GRCh38')
        self.assertEqual(type(o_vcf_2_fhir), vcf2fhir.Converter)

    def test_conv_region_only(self):
        conv_region_filename = os.path.join(os.path.dirname(
            __file__), 'RegionsToConvert_example3.bed')
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example3.vcf'),
            'GRCh37',
            'abc',
            conv_region_filename=conv_region_filename)
        self.assertEqual(type(o_vcf_2_fhir), vcf2fhir.Converter)

    def test_conv_region_dict(self):
        conv_region_dict = {
            "Chromosome": ["X", "X", "M"],
            "Start": [50000, 55000, 50000],
            "End": [52000, 60600, 60025]
        }
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example3.vcf'),
            'GRCh37',
            'abc',
            conv_region_dict=conv_region_dict)
        self.assertEqual(type(o_vcf_2_fhir), vcf2fhir.Converter)

    def test_conv_region_region_studied(self):
        region_studied_filename = os.path.join(
            os.path.dirname(__file__), 'RegionsStudied_example3.bed')
        conv_region_filename = os.path.join(os.path.dirname(
            __file__), 'RegionsToConvert_example3.bed')
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example3.vcf'),
            'GRCh37',
            'abc',
            conv_region_filename=conv_region_filename,
            region_studied_filename=region_studied_filename)
        self.assertEqual(type(o_vcf_2_fhir), vcf2fhir.Converter)

    def test_conv_region_nocall(self):
        conv_region_filename = os.path.join(os.path.dirname(
            __file__), 'RegionsToConvert_example3.bed')
        nocall_filename = os.path.join(os.path.dirname(
            __file__), 'NoncallableRegions_example3.bed')
        with self.assertRaises(Exception) as context:
            vcf2fhir.Converter(
                os.path.join(
                    os.path.dirname(__file__),
                    'vcf_example3.vcf'),
                'GRCh37',
                'abc',
                conv_region_filename=conv_region_filename,
                nocall_filename=nocall_filename)
        self.assertEqual(
            ('Please also provide region_studied_filename ' +
             'when nocall_filename is provided'), str(
                context.exception))

    def test_no_conv_region_region_studied(self):
        region_studied_filename = os.path.join(
            os.path.dirname(__file__), 'RegionsStudied_example3.bed')
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example3.vcf'),
            'GRCh37',
            'abc',
            region_studied_filename=region_studied_filename)
        self.assertEqual(type(o_vcf_2_fhir), vcf2fhir.Converter)

    def test_no_conv_region_nocall(self):
        nocall_filename = os.path.join(os.path.dirname(
            __file__), 'NoncallableRegions_example3.bed')
        with self.assertRaises(Exception) as context:
            vcf2fhir.Converter(
                os.path.join(
                    os.path.dirname(__file__),
                    'vcf_example3.vcf'),
                'GRCh37',
                'abc',
                nocall_filename=nocall_filename)
        self.assertEqual(
            ('Please also provide region_studied_filename ' +
             'when nocall_filename is provided'), str(
                context.exception))


class TestTranslation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.TEST_RESULT_DIR = os.path.join(
            os.path.dirname(__file__), 'output')
        if os.path.exists(self.TEST_RESULT_DIR):
            shutil.rmtree(self.TEST_RESULT_DIR)
        os.mkdir(self.TEST_RESULT_DIR)

    @classmethod
    def tearDownClass(self):
        shutil.rmtree(self.TEST_RESULT_DIR)

    def test_wo_patient_id(self):
        self.maxDiff = None
        o_vcf_2_fhir = vcf2fhir.Converter(os.path.join(
            os.path.dirname(__file__), 'vcf_example1.vcf'), 'GRCh37')
        output_filename = os.path.join(os.path.dirname(
            __file__), self.TEST_RESULT_DIR, 'fhir_wo_patient_example1.json')
        expected_output_filename = os.path.join(
            os.path.dirname(__file__), 'expected_example1_wo_patient.json')
        o_vcf_2_fhir.convert(output_filename)
        _compare_actual_and_expected_fhir_json(
            self,
            output_filename=output_filename,
            expected_output_filename=expected_output_filename,
            dict={8: [5, 6]})

    def test_with_patient_id(self):
        self.maxDiff = None
        o_vcf_2_fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(
            __file__), 'vcf_example1.vcf'), 'GRCh37', 'HG00628')
        output_filename = os.path.join(os.path.dirname(
            __file__), self.TEST_RESULT_DIR, 'fhir_with_patient_example1.json')
        expected_output_filename = os.path.join(os.path.dirname(
            __file__), 'expected_example1_with_patient.json')
        o_vcf_2_fhir.convert(output_filename)
        _compare_actual_and_expected_fhir_json(
            self,
            output_filename=output_filename,
            expected_output_filename=expected_output_filename,
            dict={8: [5, 6]})

    # FIXME: just a temporary test, later change it to a test that test
    # particular variant
    def test_anotherfile(self):
        o_vcf_2_fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(
            __file__), 'vcf_example2.vcf'), 'GRCh37', 'HG00628')
        o_vcf_2_fhir.convert(output_filename=os.path.join(
            os.path.dirname(__file__), self.TEST_RESULT_DIR, 'fhir2.json'))

    def test_region_studied(self):
        self.maxDiff = None
        region_studied_filename = os.path.join(
            os.path.dirname(__file__), 'RegionsStudied_example3.bed')
        conv_region_filename = os.path.join(os.path.dirname(
            __file__), 'RegionsToConvert_example3.bed')
        nocall_filename = os.path.join(os.path.dirname(
            __file__), 'NoncallableRegions_example3.bed')
        output_filename = os.path.join(os.path.dirname(
            __file__), self.TEST_RESULT_DIR, 'fhir_example3.json')
        expected_output_filename = os.path.join(
            os.path.dirname(__file__), 'expected_example3.json')
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example3.vcf'),
            'GRCh38',
            'HG00628',
            conv_region_filename=conv_region_filename,
            region_studied_filename=region_studied_filename,
            nocall_filename=nocall_filename)
        o_vcf_2_fhir.convert(output_filename)
        _compare_actual_and_expected_fhir_json(
            self,
            output_filename=output_filename,
            expected_output_filename=expected_output_filename,
            dict={7: [2, 3], 8: [3, 4]})

    def test_region_studied_dict(self):
        conv_region_dict = {
            "Chromosome": ["X", "X", "M"],
            "Start": [50000, 55000, 50000],
            "End": [52000, 60600, 60025]
        }
        self.maxDiff = None
        region_studied_filename = os.path.join(
            os.path.dirname(__file__), 'RegionsStudied_example3.bed')
        nocall_filename = os.path.join(os.path.dirname(
            __file__), 'NoncallableRegions_example3.bed')
        output_filename = os.path.join(os.path.dirname(
            __file__), self.TEST_RESULT_DIR, 'fhir_example3_dict.json')
        expected_output_filename = os.path.join(
            os.path.dirname(__file__), 'expected_example3.json')
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example3.vcf'),
            'GRCh38',
            'HG00628',
            conv_region_dict=conv_region_dict,
            region_studied_filename=region_studied_filename,
            nocall_filename=nocall_filename)
        o_vcf_2_fhir.convert(output_filename)
        _compare_actual_and_expected_fhir_json(
            self,
            output_filename=output_filename,
            expected_output_filename=expected_output_filename,
            dict={7: [2, 3], 8: [3, 4]})

    # Check if region studied observation outside the vcf files are also
    # included in fhir report.
    def test_multiple_region_studied(self):
        self.maxDiff = None
        region_studied_filename = os.path.join(
            os.path.dirname(__file__), 'RegionsStudied_example4.bed')
        conv_region_filename = os.path.join(os.path.dirname(
            __file__), 'RegionsToConvert_example4.bed')
        nocall_filename = os.path.join(os.path.dirname(
            __file__), 'NoncallableRegions_example4.bed')
        output_filename = os.path.join(os.path.dirname(
            __file__), self.TEST_RESULT_DIR, 'fhir_example4.json')
        expected_output_filename = os.path.join(
            os.path.dirname(__file__), 'expected_example4.json')
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example4.vcf'),
            'GRCh38',
            'HG00628',
            conv_region_filename=conv_region_filename,
            region_studied_filename=region_studied_filename,
            nocall_filename=nocall_filename)
        o_vcf_2_fhir.convert(output_filename)
        _compare_actual_and_expected_fhir_json(
            self,
            output_filename=output_filename,
            expected_output_filename=expected_output_filename,
            dict={31: [24, 25], 32: [25, 26]})

    def test_region_studied_only(self):
        region_studied_filename = os.path.join(
            os.path.dirname(__file__), 'RegionsStudied_example4.bed')
        output_filename = os.path.join(os.path.dirname(
            __file__), self.TEST_RESULT_DIR, 'fhir_example4_test.json')
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example4.vcf'),
            'GRCh38',
            'HG00628',
            region_studied_filename=region_studied_filename)
        o_vcf_2_fhir.convert(output_filename)

    def test_empty_fhir_json(self):
        conv_region_filename = os.path.join(os.path.dirname(
            __file__), 'RegionsToConvert_empty_example4.bed')
        output_filename = os.path.join(os.path.dirname(
            __file__), self.TEST_RESULT_DIR, 'fhir_example4_test.json')
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example4.vcf'),
            'GRCh38',
            'HG00628',
            conv_region_filename=conv_region_filename)
        o_vcf_2_fhir.convert(output_filename)

    def test_no_region_examined(self):
        self.maxDiff = None
        region_studied_filename = os.path.join(
            os.path.dirname(__file__), 'HG00403A_studied.bed')
        conv_region_filename = os.path.join(os.path.dirname(
            __file__), 'HG00403A_convert.bed')
        output_filename = os.path.join(os.path.dirname(
            __file__), self.TEST_RESULT_DIR, 'HG00403A.json')
        expected_output_filename = os.path.join(
            os.path.dirname(__file__), 'expected_HG00403A.json')
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'HG00403A.vcf.gz'),
            'GRCh37',
            has_tabix=True,
            conv_region_filename=conv_region_filename,
            region_studied_filename=region_studied_filename)
        o_vcf_2_fhir.convert(output_filename)
        _compare_actual_and_expected_fhir_json(
            self,
            output_filename=output_filename,
            expected_output_filename=expected_output_filename,
            is_null=True)

    def test_tabix(self):
        self.maxDiff = None
        region_studied_filename = os.path.join(
            os.path.dirname(__file__), 'RegionsStudied_example4.bed')
        conv_region_filename = os.path.join(os.path.dirname(
            __file__), 'RegionsToConvert_example4.bed')
        nocall_filename = os.path.join(os.path.dirname(
            __file__), 'NoncallableRegions_example4.bed')
        output_filename = os.path.join(os.path.dirname(
            __file__), self.TEST_RESULT_DIR, 'fhir_example4_tabix.json')
        expected_output_filename = os.path.join(
            os.path.dirname(__file__), 'expected_example4.json')
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example4.vcf.gz'),
            'GRCh38',
            'HG00628',
            has_tabix=True,
            conv_region_filename=conv_region_filename,
            region_studied_filename=region_studied_filename,
            nocall_filename=nocall_filename)
        o_vcf_2_fhir.convert(output_filename)
        _compare_actual_and_expected_fhir_json(
            self,
            output_filename=output_filename,
            expected_output_filename=expected_output_filename,
            dict={31: [24, 25], 32: [25, 26]})


class TestLogger(unittest.TestCase):
    def setUp(self):
        # create file handler and set level to debug
        self.log_general_filename = os.path.join(
            os.path.dirname(__file__), self.LOG_DIR, 'general.log')
        self.log_invalid_record_filename = os.path.join(
            os.path.dirname(__file__), self.LOG_DIR, 'invalid_record.log')
        genearl_fh = logging.FileHandler(self.log_general_filename)
        invalid_record_fh = logging.FileHandler(
            self.log_invalid_record_filename)
        genearl_fh.setLevel(logging.DEBUG)
        invalid_record_fh.setLevel(logging.DEBUG)
        # create formatter
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        # add formatter to fh
        genearl_fh.setFormatter(formatter)
        invalid_record_fh.setFormatter(formatter)
        # store as a class variable
        self.genearl_fh = genearl_fh
        self.invalid_record_fh = invalid_record_fh

    @classmethod
    def setUpClass(self):
        self.LOG_DIR = os.path.join(os.path.dirname(__file__), 'log')
        if os.path.exists(self.LOG_DIR):
            shutil.rmtree(self.LOG_DIR)
        os.mkdir(self.LOG_DIR)

    # TODO: Delete the log folder after running
    # all the tests, below method throws error becasue
    # log files are still in use.
    # @classmethod
    # def tearDownClass(self):
    #     shutil.rmtree(self.LOG_DIR)

    def test_logger_forks(self):
        region_studied_filename = os.path.join(
            os.path.dirname(__file__), 'RegionsStudied_example3.bed')
        conv_region_filename = os.path.join(os.path.dirname(
            __file__), 'RegionsToConvert_example3.bed')
        nocall_filename = os.path.join(os.path.dirname(
            __file__), 'NoncallableRegions_example3.bed')
        output_filename = os.path.join(os.path.dirname(
            __file__), self.LOG_DIR, 'logging_fhir.json')
        # create logger
        general_logger = logging.getLogger('vcf2fhir.general')
        invalid_record_logger = logging.getLogger('vcf2fhir.invalidrecord')
        general_logger.setLevel(logging.DEBUG)
        invalid_record_logger.setLevel(logging.DEBUG)
        # add ch to logger
        general_logger.addHandler(self.genearl_fh)
        invalid_record_logger.addHandler(self.invalid_record_fh)
        o_vcf_2_fhir = vcf2fhir.Converter(
            os.path.join(
                os.path.dirname(__file__),
                'vcf_example3.vcf'),
            'GRCh38',
            'HG00628',
            conv_region_filename=conv_region_filename,
            region_studied_filename=region_studied_filename,
            nocall_filename=nocall_filename)
        o_vcf_2_fhir.convert(output_filename)
        self.assertEqual(os.path.exists(self.log_general_filename), True)
        self.assertEqual(os.path.exists(
            self.log_invalid_record_filename), True)


suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestVcf2FhirInputs))
suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestTranslation))
suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestLogger))


if __name__ == '__main__':
    unittest.main()



import logging
from .gene_ref_seq import _get_ref_seq_by_chrom
from .fhir_helper import _Fhir_Helper
from .common import *

invalid_record_logger = logging.getLogger("vcf2fhir.invalidrecord")
general_logger = logging.getLogger("vcf2fhir.general")


def _valid_record(record):
    if not (validate_chrom_identifier(record.CHROM)):
        invalid_record_logger.debug(
            ("Reason: VCF CHROM is not recognized, " +
             "Record: %s, considered sample: %s"),
            record,
            record.samples[0].data)
        return False
    if(record.is_sv):
        invalid_record_logger.debug(
            ("Reason: VCF INFO.SVTYPE is present. " +
             "(Structural variants are excluded), " +
             "Record: %s, considered sample: %s"),
            record,
            record.samples[0].data)
        return False
    if(record.FILTER is not None and len(record.FILTER) != 0):
        invalid_record_logger.debug(
            ("Reason: VCF FILTER does not equal " +
             "'PASS' or '.', Record: %s, considered sample: %s"),
            record,
            record.samples[0].data)
        return False
    if(len(record.samples) == 0 or record.samples[0].gt_type is None):
        invalid_record_logger.debug(
            ("Reason: VCF FORMAT.GT is null ('./.', '.|.', '.', etc), " +
             "Record: %s, considered sample: %s"),
            record,
            record.samples[0].data)
        return False
    if(
            len(record.ALT) != 1 or
            (record.ALT[0].type != 'SNV' and record.ALT[0].type != 'MNV')):
        invalid_record_logger.debug(
            ("Reason: ALT is not a simple character string, comma-separated " +
             "character string or '.' (Rows where ALT is a token " +
             "(e.g. 'DEL') are excluded), Record: %s, considered sample: %s"),
            record,
            record.samples[0].data)
        return False
    if record.CHROM == "M" and (
        (len(
            record.samples[0].gt_alleles) == 1 and
            record.samples[0].gt_alleles[0] == "0") or len(
            record.samples[0].gt_alleles) == 2):
        invalid_record_logger.debug(
            ("Reason: Mitochondrial DNA with GT = 0 or its diploid, " +
             "Record: %s, considered sample: %s"),
            record,
            record.samples[0].data)
        return False
    return True


def _get_chrom(chrom_index):
    switcher = {
        23: 'X',
        24: 'Y',
        25: 'M'
    }
    return switcher.get(chrom_index, str(chrom_index))


def _fix_regions_chrom(region):
    if region:
        region.Chromosome = region.Chromosome.apply(
            extract_chrom_identifier)


def _add_record_variants(record, ref_seq, patientID, fhir_helper, ratio_ad_dp):
    if(_valid_record(record)):
        fhir_helper.add_variant_obv(record, ref_seq, ratio_ad_dp)


def _add_region_studied(
        region_studied,
        conversion_region,
        nocall_region,
        fhir_helper,
        chrom,
        ref_seq,
        patientID):
    if((
            (region_studied and not region_studied[chrom].empty) or
            (nocall_region and not nocall_region[chrom].empty)) or
            (
                (region_studied is not None and len(region_studied) == 0) and
                (conversion_region and not conversion_region[chrom].empty))):
        general_logger.info("Adding region Studied observation for %s", chrom)
        general_logger.debug("Region Examined %s", region_studied[chrom])
        general_logger.debug("Region Uncallable %s", nocall_region[chrom])
        fhir_helper.add_regionstudied_obv(
            ref_seq, region_studied[chrom], nocall_region[chrom])


def _get_fhir_json(
        vcf_reader,
        ref_build,
        patientID,
        has_tabix,
        conversion_region,
        region_studied,
        nocall_region,
        ratio_ad_dp,
        output_filename):
    fhir_helper = _Fhir_Helper(patientID)
    fhir_helper.initalize_report()
    general_logger.debug("Finished Initializing empty report")
    _fix_regions_chrom(conversion_region)
    _fix_regions_chrom(region_studied)
    _fix_regions_chrom(nocall_region)
    if conversion_region:
        if region_studied:
            region_studied = region_studied.intersect(conversion_region)
            general_logger.debug(
                "Final Conmputed Reportable Query Regions: %s", region_studied)
        if nocall_region:
            nocall_region = nocall_region.intersect(conversion_region)
            general_logger.debug(
                "Final Conmputed NoCall Query Regions: %s", nocall_region)
    general_logger.info(
        ("Start adding Region studied observation followed by " +
         "Variant observation corresponding to that region"))
    if has_tabix:
        for chrom_index in range(1, 26):
            chrom = _get_chrom(chrom_index)
            ref_seq = _get_ref_seq_by_chrom(
                ref_build, extract_chrom_identifier(chrom))
            _add_region_studied(
                region_studied,
                conversion_region,
                nocall_region,
                fhir_helper,
                chrom,
                ref_seq,
                patientID
            )
            if conversion_region and not conversion_region[chrom].empty:
                for _, row in conversion_region[chrom].df.iterrows():
                    vcf_iterator = None
                    try:
                        vcf_iterator = vcf_reader.fetch(
                            chrom, int(row['Start']), int(row['End']))
                    except ValueError:
                        pass
                    if vcf_iterator:
                        for record in vcf_iterator:
                            record.CHROM = extract_chrom_identifier(
                                record.CHROM)
                            _add_record_variants(
                                record,
                                ref_seq,
                                patientID,
                                fhir_helper,
                                ratio_ad_dp
                            )
            elif not conversion_region:
                vcf_iterator = None
                try:
                    vcf_iterator = vcf_reader.fetch(chrom)
                except ValueError:
                    pass
                if vcf_iterator:
                    for record in vcf_iterator:
                        record.CHROM = extract_chrom_identifier(
                            record.CHROM)
                        _add_record_variants(
                            record,
                            ref_seq,
                            patientID,
                            fhir_helper,
                            ratio_ad_dp
                        )
    else:
        chrom_index = 1
        prev_add_chrom = ""
        for record in vcf_reader:
            record.CHROM = extract_chrom_identifier(record.CHROM)
            if not (
                    conversion_region and
                    conversion_region[record.CHROM].empty):
                if prev_add_chrom != record.CHROM and (
                        region_studied or nocall_region):
                    chrom = _get_chrom(chrom_index)
                    while prev_add_chrom != record.CHROM:
                        current_ref_seq = _get_ref_seq_by_chrom(
                            ref_build, chrom)
                        _add_region_studied(
                            region_studied,
                            conversion_region,
                            nocall_region,
                            fhir_helper,
                            chrom,
                            current_ref_seq,
                            patientID)
                        prev_add_chrom = chrom
                        chrom_index += 1
                        chrom = _get_chrom(chrom_index)
                ref_seq = _get_ref_seq_by_chrom(ref_build, record.CHROM)
                if (
                        not conversion_region or
                        conversion_region[
                            record.CHROM,
                            record.POS - 1: (record.POS + len(record.REF) - 1)
                        ].empty is False):
                    _add_record_variants(
                        record, ref_seq, patientID, fhir_helper, ratio_ad_dp)

    general_logger.info("Adding all the phased sequence relationship found")
    fhir_helper.add_phased_relationship_obv()

    general_logger.info("Adding all the observation Ids to result block")
    fhir_helper.add_report_result()

    fhir_helper.generate_final_json()
    general_logger.info(
        f"Export the FHIR object to the file {output_filename}")
    fhir_helper.export_fhir_json(output_filename)
    general_logger.info("Completed conversion")



import doctest
import unittest
import vcf2fhir
from vcf2fhir.common import *

suite = doctest.DocTestSuite(vcf2fhir)


class TestChromIdentifier(unittest.TestCase):

    def test_chrom_1_22(self):
        actual_chrom = ['chr1', '1', 'CHR1', '22', 'CHR22', 'chr22']
        expected_chrom = ['1', '1', '1', '22', '22', '22']
        i = 0
        for chrom in actual_chrom:
            self.assertEqual(extract_chrom_identifier(
                chrom), expected_chrom[i])
            i += 1

    def test_chrom_X_Y(self):
        actual_chrom = ['chrX', 'X', 'x', 'CHRX', 'chrY', 'Y', 'Y', 'CHRY']
        expected_chrom = ['X', 'X', 'X', 'X', 'Y', 'Y', 'Y', 'Y']
        i = 0
        for chrom in actual_chrom:
            self.assertEqual(extract_chrom_identifier(
                chrom), expected_chrom[i])
            i += 1

    def test_chrom_M(self):
        actual_chrom = ['MT', 'm', 'M', 'mt', 'chrm', 'chrM', 'chrmt', 'chrMT']
        expected_chrom = ['M', 'M', 'M', 'M', 'M', 'M', 'M', 'M']
        i = 0
        for chrom in actual_chrom:
            self.assertEqual(extract_chrom_identifier(
                chrom), expected_chrom[i])
            i += 1

    def test_chrom_validation(self):
        actual_chrom = [
            'chrX',
            'R',
            'mt',
            'chrR',
            'chrM',
            'chrMT',
            'chr30',
            'P',
            'Z',
            '45',
            'o',
            'CHRX',
            '1',
            '22',
            'CHR21',
            'x']
        recognized = [
            True,
            False,
            True,
            False,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            True,
            True,
            True,
            True,
            True]
        i = 0
        for chrom in actual_chrom:
            self.assertEqual(validate_chrom_identifier(
                chrom), recognized[i])
            i += 1


class TestDataType(unittest.TestCase):

    def test_validate_ratio_ad_dp(self):
        ratio_ad_dp = [0.2, 0.65, 0.99, 1.3, -0.7, "Chr", False, True]
        valid = [True, True, True, False, False, False, False, False]
        i = 0
        for value in ratio_ad_dp:
            self.assertEqual(validate_ratio_ad_dp(
                value), valid[i])
            i += 1

    def test_validate_has_tabix(self):
        has_tabix = [True, False, 1, 23.4, 'C', "CHROM"]
        valid = [True, True, False, False, False, False]
        i = 0
        for value in has_tabix:
            self.assertEqual(validate_has_tabix(
                value), valid[i])
            i += 1


suite.addTests(unittest.TestLoader(
).loadTestsFromTestCase(TestChromIdentifier))
suite.addTests(unittest.TestLoader(
).loadTestsFromTestCase(TestDataType))

if __name__ == '__main__':
    unittest.main()



import os
import datetime
import pandas as pd
import pytz
import logging
import re
from collections import OrderedDict


general_logger = logging.getLogger("vcf2fhir.general")


"""
/**
 * generates current date into fhir format
 * @return date
 */
"""


def get_fhir_date():
    z = datetime.datetime.now(pytz.timezone(
        'UTC')).strftime("%Y-%m-%dT%H:%M:%S%z")
    return z[:-2] + ':' + z[-2:]


def get_sequence_relation(phased_rec_map):
    Relation_table = pd.DataFrame(columns=['POS1', 'POS2', 'Relation'])
    for key in phased_rec_map:
        prev_record = None
        for record in phased_rec_map[key]:
            if prev_record is None:
                prev_record = record
                continue
            prev_data = prev_record.samples[0].data
            record_data = record.samples[0].data
            if(prev_data.PS == record_data.PS):
                if prev_data.GT == record_data.GT:
                    Relation_table = Relation_table.append(
                        {
                            'POS1': prev_record.POS,
                            'POS2': record.POS,
                            'Relation': 'Cis'
                        },
                        ignore_index=True
                    )
                else:
                    Relation_table = Relation_table.append(
                        {
                            'POS1': prev_record.POS,
                            'POS2': record.POS,
                            'Relation': 'Trans'
                        },
                        ignore_index=True
                    )
            prev_record = record
    return Relation_table


def get_allelic_state(record, ratio_ad_dp):
    allelic_state = ''
    allelic_code = ''
    # Using  the first sample
    sample = record.samples[0]
    alleles = sample.gt_alleles
    if record.CHROM != 'M':
        if len(alleles) >= 2 and sample.gt_type == 1:
            allelic_state = 'heterozygous'
            allelic_code = 'LA6706-1'
        elif len(alleles) >= 2:
            allelic_state = 'homozygous'
            allelic_code = 'LA6705-3'
        elif sample.gt_type is not None and len(alleles) == 1:
            allelic_state = 'hemizygous'
            allelic_code = 'LA6707-9'
        else:
            _error_log_allelicstate(record)
    elif (
            sample.gt_type is not None and
            len(alleles) == 1 and
            alleles[0] == '1'):
        if hasattr(sample.data, 'AD') and hasattr(sample.data, 'DP'):
            try:
                if(
                        isinstance(sample.data.AD, list) and
                        len(sample.data.AD) > 0):
                    ratio = float(
                        sample.data.AD[0]) / float(sample.data.DP)
                else:
                    ratio = float(sample.data.AD) / float(sample.data.DP)
                if ratio > ratio_ad_dp:
                    allelic_state = "homoplasmic"
                    allelic_code = "LA6704-6"
                else:
                    allelic_state = "heteroplasmic"
                    allelic_code = "LA6703-8"
            except Exception as e:
                general_logger.debug(e)
                _error_log_allelicstate(record)
                pass
        else:
            _error_log_allelicstate(record)
    else:
        _error_log_allelicstate(record)
    return {'ALLELE': allelic_state, 'CODE': allelic_code}


def extract_chrom_identifier(chrom):
    chrom = chrom.upper().replace("CHR", "")
    if chrom == "MT":
        chrom = "M"
    return chrom


def validate_chrom_identifier(chrom):
    chrom = extract_chrom_identifier(chrom)
    pattern = '^[1-9]$|^1[0-9]$|^2[0-2]$|^[XYM]$'
    result = re.match(pattern, chrom)
    return bool(result)


def validate_ratio_ad_dp(ratio_ad_dp):
    if not (ratio_ad_dp):
        return False
    if not isinstance(ratio_ad_dp, float):
        return False
    if ratio_ad_dp < 0 or ratio_ad_dp >= 1:
        return False
    return True


def validate_has_tabix(has_tabix):
    if not isinstance(has_tabix, bool):
        return False
    return True


def _error_log_allelicstate(record):
    general_logger.error(
        "Cannot Determine AllelicState for: %s , considered sample: %s",
        record,
        record.samples[0].data)


def createOrderedDict(value_from, order):
    value_to = OrderedDict()
    for key in order:
        if key in value_from.keys():
            value_to[key] = value_from[key]
    return value_to



import os
import json
import fhirclient.models.diagnosticreport as dr
import fhirclient.models.codeableconcept as concept
import fhirclient.models.meta as meta
import fhirclient.models.resource as resource
import fhirclient.models.observation as observation
import fhirclient.models.fhirreference as reference
import fhirclient.models.fhirdate as date
import fhirclient.models.range as valRange
import fhirclient.models.medicationstatement as medication
import numpy as np
from uuid import uuid4
from .common import *

CG_ORDER = ["system", "code"]
CODE_ORD = ["system", "code", "display"]
RS_ORDER = ['resourceType', 'id', 'meta', 'status', 'category', 'code',
            'subject', 'component']
DV_ORDER = ['resourceType', 'id', 'meta', 'status', 'category', 'code',
            'subject', 'valueCodeableConcept', 'component']
SID_ORDER = ['resourceType', 'id', 'meta', 'status', 'category',
             'code', 'subject', 'valueCodeableConcept', 'derivedFrom']


class _Fhir_Helper:
    def __init__(self, patientID):
        self.report = dr.DiagnosticReport()
        self.phased_rec_map = {}
        self.result_ids = []
        self.fhir_report = {}
        self.obs_contained = []
        self.patientID = patientID

    def _get_region_studied_component(
            self,
            reportable_query_regions,
            nocall_regions):
        observation_rs_components = []
        ranges_examined_component = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "51959-5",
                        "display": "Range(s) of DNA sequence examined"
                    }
                ]
            }
        )
        if (
                reportable_query_regions is not None and
                len(reportable_query_regions) == 0):
            obv_comp = observation.ObservationComponent()
            obv_comp.code = ranges_examined_component
            obv_comp\
                .dataAbsentReason = concept.CodeableConcept(
                    {
                        "coding": [
                            {
                                "system": ("http://terminology.hl7.org/" +
                                           "CodeSystem/data-absent-reason"),
                                "code": "not-performed",
                                "display": "Not Performed"
                            }
                        ]
                    }
                )
            observation_rs_components.append(obv_comp)
            return observation_rs_components
        for _, row in reportable_query_regions.df.iterrows():
            obv_comp = observation.ObservationComponent()
            obv_comp.code = ranges_examined_component
            obv_comp.valueRange = valRange.Range({"low": {"value": float(
                row['Start']) + 1},
                "high": {"value": float(row['End']) + 1}})
            observation_rs_components.append(obv_comp)
        for _, row in nocall_regions.df.iterrows():
            obv_comp = observation.ObservationComponent()
            obv_comp.code = concept.CodeableConcept(
                {
                    "coding": [
                        {
                            "system": ("http://hl7.org/fhir/uv/genomics" +
                                       "-reporting/CodeSystem/TbdCodes"),
                            "code": "uncallable-regions",
                            "display": "Uncallable Regions"
                        }
                    ]
                }
            )
            obv_comp.valueRange = valRange.Range({"low": {"value": float(
                row['Start']) + 1},
                "high": {"value": float(row['End']) + 1}})
            observation_rs_components.append(obv_comp)
        return observation_rs_components

    def _add_phase_records(self, record):
        if(record.samples[0].phased is False):
            return
        sample_data = record.samples[0].data
        if(
                sample_data.GT is not None and
                len(sample_data.GT.split('|')) >= 2 and
                'PS' in sample_data._fields):
            self.phased_rec_map.setdefault(sample_data.PS, []).append(record)

    def initalize_report(self):
        patient_reference = reference.FHIRReference(
            {"reference": "Patient/" + self.patientID})
        self.report.id = "dr-" + uuid4().hex[:13]
        self.report.meta = meta.Meta({
                            "profile": [
                                ("http://hl7.org/fhir/uv/genomics-reporting" +
                                 "/StructureDefinition/genomics-report")]})
        self.report.status = "final"
        self.report.category = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": ("http://terminology.hl7.org/" +
                                   "CodeSystem/v2-0074"),
                        "code": "GE"
                    }
                ]
            }
        )
        self.report.code = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "81247-9",
                        "display": "Master HL7 genetic variant reporting panel"
                    }
                ]
            }
        )
        self.report.subject = patient_reference
        self.report.issued = date.FHIRDate(get_fhir_date())
        self.report.contained = []

    def add_regionstudied_obv(
            self,
            ref_seq,
            reportable_query_regions,
            nocall_regions):
        if(((
                not (reportable_query_regions is not None) and
                (len(reportable_query_regions) == 0))) and
                (reportable_query_regions.empty and nocall_regions.empty)):
            return
        patient_reference = reference.FHIRReference(
            {"reference": "Patient/" + self.patientID})
        contained_uid = "rs-" + uuid4().hex[:13]
        self.result_ids.append(contained_uid)
        # Region Studied Obeservation
        observation_rs = observation.Observation()
        contained_rs = observation_rs
        contained_rs.id = contained_uid
        observation_rs.resource_type = "Observation"
        contained_rs.meta = meta.Meta({"profile": [
                                      ("http://hl7.org/fhir/uv/" +
                                       "genomics-reporting/" +
                                       "StructureDefinition/region-studied")]})
        observation_rs.code = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "53041-0",
                        "display": "DNA region of interest panel"
                    }
                ]
            }
        )
        observation_rs.status = "final"
        observation_rs.category = [concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": ("http://terminology.hl7.org/" +
                                   "CodeSystem/observation-category"),
                        "code": "laboratory"
                    }
                ]
            }
        )]
        observation_rs.subject = patient_reference
        observation_rs_component1 = observation.ObservationComponent()
        observation_rs_component1.code = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "92822-6",
                        "display": "Genomic coord system"
                    }
                ]
            }
        )
        observation_rs_component1\
            .valueCodeableConcept = concept.CodeableConcept(
                {
                    "coding": [
                        {
                            "system": "http://loinc.org",
                            "code": "LA30102-0",
                            "display": "1-based character counting"
                        }
                    ]
                }
            )
        observation_rs_component2 = observation.ObservationComponent()
        observation_rs_component2.code = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "48013-7",
                        "display": "Genomic reference sequence ID"
                    }
                ]
            }
        )
        observation_rs_component2\
            .valueCodeableConcept = concept.CodeableConcept(
                {
                    "coding": [
                        {
                            "system": "http://www.ncbi.nlm.nih.gov/nuccore",
                            "code": ref_seq
                        }
                    ]
                }
            )
        observation_rs_components = self._get_region_studied_component(
            reportable_query_regions, nocall_regions)
        observation_rs.component = [
            observation_rs_component1,
            observation_rs_component2] + observation_rs_components
        # Observation structure : described-variants
        self.report.contained.append(contained_rs)

    def add_variant_obv(self, record, ref_seq, ratio_ad_dp):
        # collect all the record with similar position values,
        # to utilized later in phased sequence relationship
        self._add_phase_records(record)
        patient_reference = reference.FHIRReference(
            {"reference": "Patient/" + self.patientID})
        alleles = get_allelic_state(record, ratio_ad_dp)
        dvuid = "dv-" + uuid4().hex[:13]
        self.fhir_report.update({str(record.POS): dvuid})
        self.result_ids.append(dvuid)
        observation_dv = observation.Observation()
        observation_dv.resource_type = "Observation"
        observation_dv.id = dvuid
        observation_dv.meta = meta.Meta(
            {
                "profile": [
                    ("http://hl7.org/fhir/uv/" +
                     "genomics-reporting/StructureDefinition/variant")
                ]
            }
        )
        observation_dv.status = "final"
        observation_dv.category = [concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": ("http://terminology.hl7.org/" +
                                   "CodeSystem/observation-category"),
                        "code": "laboratory"
                    }
                ]
            }
        )]
        observation_dv.code = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "69548-6",
                        "display": "Genetic variant assessment"
                    }
                ]
            }
        )
        observation_dv.subject = patient_reference
        observation_dv.valueCodeableConcept = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "LA9633-4",
                        "display": "present"
                    }
                ]
            }
        )
        observation_dv.component = []

        observation_dv_component2 = observation.ObservationComponent()
        observation_dv_component2.code = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "48013-7",
                        "display": "Genomic reference sequence ID"
                    }
                ]
            }
        )
        observation_dv_component2\
            .valueCodeableConcept = concept.CodeableConcept(
                {
                    "coding": [
                        {
                            "system": "http://www.ncbi.nlm.nih.gov/nuccore",
                            "code": ref_seq
                        }
                    ]
                }
            )
        observation_dv.component.append(observation_dv_component2)

        if alleles['CODE'] != "" or alleles['ALLELE'] != "":
            observation_dv_component3 = observation.ObservationComponent()
            observation_dv_component3.code = concept.CodeableConcept(
                {
                    "coding": [
                        {
                            "system": "http://loinc.org",
                            "code": "53034-5",
                            "display": "Allelic state"
                        }
                    ]
                }
            )
            observation_dv_component3\
                .valueCodeableConcept = concept.CodeableConcept(
                    {
                        "coding": [
                            {
                                "system": "http://loinc.org",
                                "code": alleles['CODE'],
                                "display": alleles['ALLELE']
                            }
                        ]
                    }
                )
            observation_dv.component.append(observation_dv_component3)

        observation_dv_component4 = observation.ObservationComponent()
        observation_dv_component4.code = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "69547-8",
                        "display": "Genomic Ref allele [ID]"
                    }
                ]
            }
        )
        observation_dv_component4.valueString = record.REF
        observation_dv.component.append(observation_dv_component4)

        observation_dv_component5 = observation.ObservationComponent()
        observation_dv_component5.code = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "69551-0",
                        "display": "Genomic Alt allele [ID]"
                    }
                ]
            }
        )
        observation_dv_component5.valueString = record.ALT[0].sequence
        observation_dv.component.append(observation_dv_component5)

        observation_dv_component6 = observation.ObservationComponent()
        observation_dv_component6.code = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "92822-6",
                        "display": "Genomic coord system"
                    }
                ]
            }
        )
        observation_dv_component6\
            .valueCodeableConcept = concept.CodeableConcept(
                {
                    "coding": [
                        {
                            "system": "http://loinc.org",
                            "code": "LA30102-0",
                            "display": "1-based character counting"
                        }
                    ]
                }
            )
        observation_dv.component.append(observation_dv_component6)

        observation_dv_component7 = observation.ObservationComponent()
        observation_dv_component7.code = concept.CodeableConcept(
            {
                "coding": [
                    {
                        "system": ("http://hl7.org/fhir/uv/" +
                                   "genomics-reporting/CodeSystem/TbdCodes"),
                        "code": "exact-start-end",
                        "display": "Variant exact start and end"}]})
        observation_dv_component7.valueRange = valRange.Range(
            {"low": {"value": int(record.POS)}})
        observation_dv.component.append(observation_dv_component7)

        self.report.contained.append(observation_dv)

    def add_phased_relationship_obv(self):
        patient_reference = reference.FHIRReference(
            {"reference": "Patient/" + self.patientID})
        self.sequence_rels \
            = get_sequence_relation(self.phased_rec_map)
        for index in self.sequence_rels.index:
            siduid = "sid-" + uuid4().hex[:13]
            self.result_ids.append(siduid)

            observation_sid = observation.Observation()
            observation_sid.resource_type = "Observation"
            observation_sid.id = siduid
            observation_sid.meta = meta.Meta({"profile": [
                                             ("http://hl7.org/fhir/uv/" +
                                              "genomics-reporting/" +
                                              "StructureDefinition/" +
                                              "sequence-phase-relationship")]})
            observation_sid.status = "final"
            observation_sid.category = [concept.CodeableConcept(
                {
                    "coding": [
                        {
                            "system": ("http://terminology.hl7.org/" +
                                       "CodeSystem/observation-category"),
                            "code": "laboratory"
                        }
                    ]
                }
            )]
            observation_sid.code = concept.CodeableConcept(
                {
                    "coding": [
                        {
                            "system": "http://loinc.org",
                            "code": "82120-7",
                            "display": "Allelic phase"
                        }
                    ]
                }
            )
            observation_sid.subject = patient_reference
            observation_sid.valueCodeableConcept = concept.CodeableConcept(
                {
                    "coding": [
                        {
                            "system": ("http://hl7.org/fhir/uv/" +
                                       "genomics-reporting/CodeSystem/" +
                                       "SequencePhaseRelationshipCS"),
                            "code": self.sequence_rels.at[index, 'Relation'],
                            "display":self.sequence_rels.at[index, 'Relation']
                        }
                    ]
                }
            )
            self.report.contained.append(observation_sid)

    def add_report_result(self):
        report_result = []
        for uid in self.result_ids:
            report_result.append(
                reference.FHIRReference({"reference": f"#{uid}"}))
        self.report.result = report_result

    def generate_final_json(self):
        response = self.report.as_json()
        od = OrderedDict()
        od["resourceType"] = response['resourceType']
        od["id"] = response['id']
        od["meta"] = response['meta']
        if 'contained' in response:
            od["contained"] = response['contained']
        else:
            od["contained"] = []
        od["status"] = response['status']
        od['category'] = []
        od['category'].append(response['category'])
        od['category'][0]['coding'][0] =\
            createOrderedDict(od['category'][0]['coding'][0], CG_ORDER)
        od["code"] = response['code']
        od["subject"] = response['subject']
        od["issued"] = response['issued']
        if 'result' in response:
            od["result"] = response['result']
        else:
            od["result"] = []
        od['code']['coding'][0] =\
            createOrderedDict(od['code']['coding'][0], CODE_ORD)

        sidIndex = 0
        for index, fhirReport in enumerate(od['contained']):
            if (fhirReport['id'].startswith('sid-')):
                sidIndex = index
                break

        for index, (_, fhirReport) in enumerate(
                zip(self.sequence_rels.index,
                    od['contained'][sidIndex:])):
            dvRef1 = self.fhir_report.get(
                str(self.sequence_rels.at[index, 'POS1']))
            dvRef2 = self.fhir_report.get(
                str(self.sequence_rels.at[index, 'POS2']))
            if (fhirReport['id'].startswith('sid-')):
                derived_from_DV1 = {}
                derived_from_DV2 = {}
                derived_from_DV1['reference'] = f"#{dvRef1}"
                derived_from_DV2['reference'] = f"#{dvRef2}"
                derivedFrom = [derived_from_DV1, derived_from_DV2]
                fhirReport['derivedFrom'] = derivedFrom

        for k, i in enumerate(od['contained']):
            od_contained_k = od['contained'][k]
            v_c_c = 'valueCodeableConcept'

            if (i['category'][0]['coding'][0]):
                od_contained_k['category'][0]['coding'][0] =\
                    createOrderedDict(i['category'][0]['coding'][0], CG_ORDER)

            if (i['code']['coding'][0]):
                od_contained_k['code']['coding'][0] =\
                    createOrderedDict(i['code']['coding'][0], CODE_ORD)

            if v_c_c in i.keys():
                od_contained_k[v_c_c]['coding'][0] =\
                    createOrderedDict(i[v_c_c]['coding'][0], CODE_ORD)

            if ((i['id'].startswith('dv-')) or (i['id'].startswith('rs-'))):
                for q, j in enumerate(i['component']):
                    od_contained_k_component_q = od_contained_k['component'][q]
                    if od_contained_k_component_q['code']['coding'][0]:
                        od_contained_k_component_q['code']['coding'][0] =\
                            createOrderedDict(j['code']['coding'][0], CODE_ORD)

                    if v_c_c in j.keys():
                        od_contained_k_component_q[v_c_c]['coding'][0] =\
                            createOrderedDict(j[v_c_c]['coding'][0], CODE_ORD)

            if (i['id'].startswith('rs-')):
                od['contained'][k] = createOrderedDict(i, RS_ORDER)

            if (i['id'].startswith('dv-')):
                od['contained'][k] = createOrderedDict(i, DV_ORDER)

            if (i['id'].startswith('sid-')):
                od['contained'][k] = createOrderedDict(i, SID_ORDER)
        self.fhir_json = od

    def export_fhir_json(self, output_filename):
        with open(output_filename, 'w') as fp:
            json.dump(self.fhir_json, fp, indent=4)
